# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os, base64
import requests
import faassupervisor.utils as utils
from faassupervisor.interfaces.dataprovider import DataProviderInterface

class Onedata(DataProviderInterface):

    CDMI_PATH = '/cdmi'

    def __init__(self, event, output_folder):
        self.event = event
        self.os_tmp_folder = os.path.dirname(output_folder)
        self.output_folder = output_folder
        # Onedata settings
        self.onedata_access_token = os.environ.get('ONEDATA_ACCESS_TOKEN')
        self.oneprovider_host = os.environ.get('ONEPROVIDER_HOST')
        self.oneprovider_space = os.environ.get('ONEDATA_SPACE')
        self.headers = {
            'X-Auth-Token': self.onedata_access_token
        }

    @classmethod    
    def is_onedata_event(cls, event):
        if utils.is_key_and_value_in_dictionary('data', event) \
        and utils.is_key_and_value_in_dictionary('body', event['data']) \
        and utils.is_key_and_value_in_dictionary('eventSource', event['data']['body']):
            event_source = event['data']['body']['eventSource']
            return event_source == 'OneTrigger'
        return False

    def download_input(self):
        '''Downloads the file from the Onedata space and returns the path were the download is placed'''
        file_path = self.event['data']['body']['path']
        file_name = self.event['data']['body']['file']
        url = 'https://{0}{1}{2}'.format(self.oneprovider_host, self.CDMI_PATH, file_path)
        print("Downloading item from '{0}' with key '{1}'".format(file_path, file_name))
        req = requests.get(url, headers=self.headers)
        if req.status_code == 200:
            file_download_path = "{0}/{1}".format(self.os_tmp_folder, file_name) 
            utils.create_folder(self.os_tmp_folder)
            print(file_name)
            with open(file_download_path, 'wb') as f:
                f.write(req.content)
            print("Successful download of file '{0}' from Oneprovider '{1}' in path '{2}'".format(file_name, file_path, file_download_path))
            return file_download_path
        else:
            print("Download failed!")
            return None

    def upload_output(self):
        output_files_path = utils.get_all_files_in_directory(self.output_folder)
        print("UPLOADING FILES {0}".format(output_files_path))
        for file_path in output_files_path:
            file_name = os.path.basename(file_path)
            output_file_name = "{0}-out{1}".format(os.path.splitext(file_name)[0],''.join(os.path.splitext(file_name)[1:]))
            self.upload_file(file_path, output_file_name)

    def upload_file(self, file_path, file_name):
        # Get OUTPUT_BUCKET environment variable to use as output folder in Onedata Space
        # If not set upload files to space root
        output_bucket = os.environ.get('OUTPUT_BUCKET')
        if output_bucket == None:
            url = 'https://{0}{1}/{2}/{3}'.format(self.oneprovider_host, self.CDMI_PATH, self.oneprovider_space, file_name)
        else:
            url = 'https://{0}{1}/{2}/{3}/{4}'.format(self.oneprovider_host, self.CDMI_PATH, self.oneprovider_space, output_bucket, file_name)
        print("Uploading file  '{0}' to '{1}/{2}'".format(file_name, self.oneprovider_space, output_bucket))
        with open(file_path, 'rb') as f:
            req = requests.put(url, data=f, headers=self.headers)
        if req.status_code not in [201, 202, 204]:
            print("Upload failed. Status code: {0}".format(req.status_code))


# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import subprocess
import faassupervisor.utils as utils
import faassupervisor.exceptions as excp
from faassupervisor.interfaces.supervisor import SupervisorInterface
from faassupervisor.providers.onpremises.storage.minio import Minio
from faassupervisor.providers.onpremises.storage.onedata import Onedata

logger = utils.get_logger()
logger.info('SUPERVISOR: Initializing Openfaas supervisor')

class OpenfaasSupervisor(SupervisorInterface):
    
    # output_folder = utils.join_paths(utils.get_random_tmp_folder(), "output")
    
    def __init__(self, **kwargs):
        logger.info('SUPERVISOR: Initializing Openfaas supervisor')
        self.event = kwargs['event']
        self.output_folder = utils.create_tmp_dir()
        self.output_folder_path = utils.join_paths(self.output_folder.name, "output")
        utils.create_folder(self.output_folder_path)
        utils.set_environment_variable('SCAR_OUTPUT_FOLDER', self.output_folder_path)

    @utils.lazy_property
    def storage_client(self):
        if Minio.is_minio_event(self.event):
            storage_client = Minio(self.event, self.output_folder_path)
        elif Onedata.is_onedata_event(self.event):
            storage_client = Onedata(self.event, self.output_folder_path)
        else:
            raise excp.NoStorageProviderDefinedWarning()
        return storage_client
       
    ##################################################################
    ## The methods below must be defined for the supervisor to work ##
    ##################################################################
    
    def execute_function(self):
        if utils.is_variable_in_environment('sprocess'):
            print("Executing user_script.sh")
            print(subprocess.call(['/bin/sh', utils.get_environment_variable('sprocess')], stderr=subprocess.STDOUT))    
    
    @excp.exception(logger)    
    def parse_input(self):
        try:
            utils.set_environment_variable('SCAR_INPUT_FILE', self.storage_client.download_input())
            logger.info('SCAR_INPUT_FILE: {0}'.format(utils.get_environment_variable('SCAR_INPUT_FILE')))
        except excp.NoStorageProviderDefinedWarning:
            pass
    
    @excp.exception(logger)
    def parse_output(self):
        try:        
            self.storage_client.upload_output()
        except excp.NoStorageProviderDefinedWarning:
            pass

    def create_response(self):
        pass
    
    def create_error_response(self):
        pass



# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import json
import faassupervisor.utils as utils
import faassupervisor.exceptions as excp
from faassupervisor.providers.aws.lambda_.supervisor import LambdaSupervisor
from faassupervisor.providers.aws.batch.supervisor import BatchSupervisor
from faassupervisor.providers.onpremises.openfaas.supervisor import OpenfaasSupervisor

logger = utils.get_logger()

class Supervisor():
    
    allowed_supervisor_types = ['LAMBDA', 'BATCH', 'OPENFAAS']

    def __init__(self, typ, **kwargs):
        ''' The class names initialized must follow the naming pattern 'type.lower().capitalize() + Supervisor'.
        For example the class 'LambdaSupervisor' is: 'LAMBDA'.lower().capitalize() + 'Supervisor'.
        '''
        targetclass = "{0}{1}".format(typ.lower().capitalize(), 'Supervisor')
        self.supervisor =  globals()[targetclass](**kwargs)

    def run(self):
        try:
            self.supervisor.parse_input()
            self.supervisor.execute_function()
            self.supervisor.parse_output()
        except Exception:
            return self.supervisor.create_error_response()
        logger.info('Creating response')
        return self.supervisor.create_response()
    
def _get_supervisor_type():
    typ = utils.get_environment_variable("SUPERVISOR_TYPE")
    _is_allowed_environment(typ)
    return typ

@excp.exception(logger)
def _is_allowed_environment(typ):
    if typ not in Supervisor.allowed_supervisor_types:
        raise excp.InvalidSupervisorTypeError(sup_typ=typ)

def _get_stdin():
    buf = ""
    for line in sys.stdin:
        buf = buf + line
    return buf

def _start_supervisor(**kwargs):
    typ = _get_supervisor_type()
    supervisor = Supervisor(typ, **kwargs)
    return supervisor.run()

def python_main(**kwargs):
    ''' Called when running from a Python environment.
    Receives the input from the method arguments.
    '''
    return _start_supervisor(**kwargs)

def main():
    ''' Called when running as binary.
    Receives the input from stdin.
    '''
    kwargs = {'event': _get_stdin()}
    return _start_supervisor(**kwargs)
    
if __name__ == "__main__":
    main()



# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from botocore.exceptions import ClientError
import functools
import sys

def exception(logger):
    '''
    A decorator that wraps the passed in function and logs exceptions
    @param logger: The logging object
    '''
    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except ClientError as ce:
                print("There was an exception in {0}".format(func.__name__))
                print(ce.response['Error']['Message'])
                logger.exception(ce)
                sys.exit(1)
            except FaasSupervisorError as fse:
                print(fse.args[0])
                logger.exception(fse)
                # Finish the execution if it's an error
                if 'Error' in fse.__class__.__name__:
                    sys.exit(1)
            except Exception as ex:
                print("There was an unmanaged exception in {0}".format(func.__name__))
                logger.exception(ex)
                sys.exit(1)
        return wrapper
    return decorator

class FaasSupervisorError(Exception):
    """
    The base exception class for exceptions.

    :ivar msg: The descriptive message associated with the error.
    """
    fmt = 'An unspecified error occurred'

    def __init__(self, **kwargs):
        msg = self.fmt.format(**kwargs)
        Exception.__init__(self, msg)
        self.kwargs = kwargs

################################################
##             GENERAL EXCEPTIONS             ##
################################################
class InvalidPlatformError(FaasSupervisorError):
    """
    The binary is not launched on a Linux platform

    """
    fmt = "This binary only works on a Linux Platform.\nTry executing the Python version."

class InvalidSupervisorTypeError(FaasSupervisorError):
    """
    The supervisor type is not in the allowed list

    """
    fmt = "The supervisor type '{sup_typ}' is not allowed."    
    
################################################
##        STORAGE PROVIDER EXCEPTIONS         ##
################################################
class NoStorageProviderDefinedWarning(FaasSupervisorError):
    """
    There is no storage provider defined.

    """
    fmt = "There is no storage provider defined for this function execution."
    
class NoInputStorageProviderDefinedWarning(NoStorageProviderDefinedWarning):
    """
    There is no input storage provider defined.

    """
    fmt = "There is no input storage provider defined for this function execution."
    
class NoOutputStorageProviderDefinedWarning(NoStorageProviderDefinedWarning):
    """
    There is no output storage provider defined.

    """
    fmt = "There is no output storage provider defined for this function execution."    
 
    


# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os
import shutil
import subprocess
import socket
import faassupervisor.utils as utils
logger = utils.get_logger()

class Udocker():

    container_name = "udocker_container"
    script_exec = "/bin/sh"

    def __init__(self, lambda_instance, scar_input_file):
        self.lambda_instance = lambda_instance
        self.scar_input_file = scar_input_file        
        
        self.udocker_exec = [utils.get_environment_variable("UDOCKER_EXEC")]
        self.container_output_file = "{}/container-stdout.txt".format(self.lambda_instance.temporal_folder_path)
        
        if utils.is_variable_in_environment("IMAGE_ID"):
            self.container_image_id = utils.get_environment_variable("IMAGE_ID")
            self.set_udocker_commands()
        else:
            raise Exception("Container image id not specified.")
    
    def set_udocker_commands(self):
        self.cmd_get_images = self.udocker_exec + ["images"]
        self.cmd_load_image = self.udocker_exec + ["load", "-i", self.container_image_id]
        self.cmd_download_image = self.udocker_exec + ["pull", self.container_image_id]
        self.cmd_list_containers = self.udocker_exec + ["ps"]
        self.cmd_create_container = self.udocker_exec + ["create", "--name={0}".format(self.container_name), self.container_image_id]
        self.cmd_set_execution_mode = self.udocker_exec + ["setup", "--execmode=F1", self.container_name]
        self.cmd_container_execution = self.udocker_exec + ["--quiet", "run"]
        
    def is_container_image_downloaded(self):
        cmd_out = utils.execute_command_and_return_output(self.cmd_get_images)
        return self.container_image_id in cmd_out

    def create_image(self):
        if self.is_container_image_downloaded():
            logger.info("Container image '{0}' already available".format(self.container_image_id))
        else:
            if utils.is_variable_in_environment("IMAGE_FILE"):
                self.load_local_container_image()
            else:
                self.download_container_image()

    def load_local_container_image(self):
        logger.info("Loading container image '{0}'".format(self.container_image_id))
        utils.execute_command(self.cmd_load_image)
        
    def download_container_image(self):
        logger.info("Pulling container '{0}' from Docker Hub".format(self.container_image_id))
        utils.execute_command(self.cmd_download_image)

    def is_container_available(self):
        cmd_out = utils.execute_command_and_return_output(self.cmd_list_containers)
        return self.container_name in cmd_out

    def create_container(self):
        if self.is_container_available():
            logger.info("Container already available")
        else:
            logger.info("Creating container based on image '{0}'.".format(self.container_image_id))
            utils.execute_command(self.cmd_create_container)
        utils.execute_command(self.cmd_set_execution_mode)

    def create_command(self):
        self.add_container_volumes()
        self.add_container_environment_variables()
        # Container running script
        if utils.is_value_in_dict(self.lambda_instance.event, 'script'): 
            # Add script in memory as entrypoint
            script_path = "{0}/script.sh".format(self.lambda_instance.temporal_folder_path)
            script_content = utils.base64_to_utf8_string(self.lambda_instance.event['script'])
            utils.create_file_with_content(script_path, script_content)
            self.cmd_container_execution += ["--entrypoint={0} {1}".format(self.script_exec, script_path), self.container_name]
        # Container with args
        elif utils.is_value_in_dict(self.lambda_instance.event,'cmd_args'):
            # Add args
            self.cmd_container_execution += [self.container_name]
            self.cmd_container_execution += json.loads(self.lambda_instance.event['cmd_args'])
        # Script to be executed every time (if defined)
        elif utils.is_variable_in_environment('INIT_SCRIPT_PATH'):
            # Add init script
            init_script_path = "{0}/init_script.sh".format(self.lambda_instance.temporal_folder_path)
            shutil.copyfile(utils.get_environment_variable("INIT_SCRIPT_PATH"), init_script_path)    
            self.cmd_container_execution += ["--entrypoint={0} {1}".format(self.script_exec, init_script_path), self.container_name]
        # Only container
        else:
            self.cmd_container_execution += [self.container_name]
    
    def add_container_volumes(self):
        self.cmd_container_execution += ["-v", self.lambda_instance.temporal_folder_path]
        self.cmd_container_execution += ["-v", "/dev", "-v", "/proc", "-v", "/etc/hosts", "--nosysdirs"]
        if utils.is_variable_in_environment('EXTRA_PAYLOAD'):
            self.cmd_container_execution += ["-v", self.lambda_instance.permanent_folder]

    def add_container_environment_variable(self, key, value):
        self.cmd_container_execution += self.parse_container_environment_variable(key, value)
            
    def add_container_environment_variables(self):
        self.cmd_container_execution += self.parse_container_environment_variable("REQUEST_ID", self.lambda_instance.request_id)
        self.cmd_container_execution += self.parse_container_environment_variable("INSTANCE_IP", 
                                                                                  socket.gethostbyname(socket.gethostname()))        
        self.cmd_container_execution += self.get_user_defined_variables()
        self.cmd_container_execution += self.get_iam_credentials()        
        self.cmd_container_execution += self.get_input_file()
        self.cmd_container_execution += self.get_output_dir()
        self.cmd_container_execution += self.get_extra_payload_path()
        self.cmd_container_execution += self.get_lambda_output_variable()
       
    def parse_container_environment_variable(self, key, value):
        var = []
        if key and value and key != "" and value != "":
            var += ["--env", str(key) + '=' + str(value)]
        return var
        
    def get_user_defined_variables(self):
        result = []
        for key,value in utils.get_user_defined_variables().items():
            result += self.parse_container_environment_variable(key, value)
        if hasattr(self.lambda_instance, "http_params") and self.lambda_instance.http_params:
            for key,value in self.lambda_instance.http_params.items():
                result += self.parse_container_environment_variable(key, value)
        return result

    def get_iam_credentials(self):
        credentials = []
        iam_creds = {'CONT_VAR_AWS_ACCESS_KEY_ID':'AWS_ACCESS_KEY_ID',
                     'CONT_VAR_AWS_SECRET_ACCESS_KEY':'AWS_SECRET_ACCESS_KEY',
                     'CONT_VAR_AWS_SESSION_TOKEN':'AWS_SESSION_TOKEN'}
        # Add IAM credentials
        for key,value in iam_creds.items():
            if not utils.is_variable_in_environment(key):
                credentials += self.parse_container_environment_variable(value, utils.get_environment_variable(value))
        return credentials
    
    def get_input_file(self):
        file = []
        if self.scar_input_file and self.scar_input_file != "":
            file += self.parse_container_environment_variable("SCAR_INPUT_FILE", self.scar_input_file)
        return file
    
    def get_output_dir(self):
        return self.parse_container_environment_variable("SCAR_OUTPUT_DIR", 
                                                         "{}".format(self.lambda_instance.output_folder))
            
    def get_extra_payload_path(self):
        ppath = []
        if utils.is_variable_in_environment('EXTRA_PAYLOAD'):
            ppath += self.parse_container_environment_variable("EXTRA_PAYLOAD", 
                                                               utils.get_environment_variable("EXTRA_PAYLOAD"))
        return ppath
          
    def get_lambda_output_variable(self):
        out_lambda = []
        if utils.is_variable_in_environment('OUTPUT_LAMBDA'):
            utils.set_environment_variable("OUTPUT_LAMBDA_FILE", "{0}/lambda_output".format(self.lambda_instance.temporal_folder_path))
            out_lambda += self.parse_container_environment_variable("OUTPUT_LAMBDA_FILE", 
                                                                    utils.get_environment_variable("EXTRA_PAYLOAD"))
        return out_lambda      
            
    def launch_udocker_container(self):
        remaining_seconds = self.lambda_instance.get_invocation_remaining_seconds()
        logger.info("Executing udocker container. Timeout set to {0} seconds".format(remaining_seconds))
        logger.debug("Udocker command: {0}".format(self.cmd_container_execution))
        with open(self.container_output_file, "w", encoding="latin-1") as out:
            with subprocess.Popen(self.cmd_container_execution, 
                                  stderr=subprocess.STDOUT, 
                                  stdout=out, 
                                  preexec_fn=os.setsid) as process:
                try:
                    process.wait(timeout=remaining_seconds)
                except subprocess.TimeoutExpired:
                    logger.info("Stopping process '{0}'".format(process))
                    process.kill()
                    logger.warning("Container timeout")
                    raise
        if os.path.isfile(self.container_output_file):
            return utils.read_file(self.container_output_file, file_encoding="latin-1")



# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import boto3
import json
import faassupervisor.utils as utils
logger = utils.get_logger()

# Used for defining Batch jobs in the Batch environment
class Batch():
    
    @utils.lazy_property
    def client(self):
        client = boto3.client('batch')
        return client    
    
    def __init__(self, lambda_instance, scar_input_file):
        self.lambda_instance = lambda_instance
        self.scar_batch_io_image_id = utils.get_environment_variable('BATCH_SUPERVISOR_IMG')
        self.script = self.get_user_script()
        self.scar_input_file = scar_input_file
        self.io_job_name = "{0}-io".format(lambda_instance.function_name)
        self.container_environment_variables = []
        self.create_context()
    
    def create_context(self):
        self.context = {'function_name': self.lambda_instance.context.function_name,
                        'memory_limit_in_mb': self.lambda_instance.context.memory_limit_in_mb,
                        'aws_request_id': self.lambda_instance.context.aws_request_id,
                        'log_group_name': self.lambda_instance.context.log_group_name,
                        'log_stream_name': self.lambda_instance.context.log_stream_name}
    
    def set_container_variables(self, step):
        self.add_environment_variable("STEP", step)
        self.add_environment_variable("SCRIPT", self.script)
        self.add_environment_variable("FUNCTION_NAME", self.lambda_instance.function_name)
        self.add_environment_variable("LAMBDA_EVENT", json.dumps(self.lambda_instance.event))
        self.add_environment_variable("LAMBDA_CONTEXT", json.dumps(self.context))
        self.add_environment_variable("SCAR_INPUT_DIR", self.lambda_instance.input_folder)
        self.add_environment_variable("SCAR_OUTPUT_DIR", self.lambda_instance.output_folder)
        self.add_environment_variable("REQUEST_ID", self.lambda_instance.request_id)

        if self.scar_input_file:
            self.add_environment_variable("SCAR_INPUT_FILE", self.scar_input_file)
        if self.lambda_instance.has_input_bucket():
            self.add_environment_variable("INPUT_BUCKET", self.lambda_instance.input_bucket)
        if self.lambda_instance.has_output_bucket():
            self.add_environment_variable("OUTPUT_BUCKET", self.lambda_instance.output_bucket)
        if self.lambda_instance.has_output_bucket_folder():
            self.add_environment_variable("OUTPUT_FOLDER", self.lambda_instance.output_bucket_folder)                 
        
        for user_var, value in utils.get_user_defined_variables().items():
            self.add_environment_variable(user_var, value)
    
    def add_environment_variable(self, name, value):
        return self.container_environment_variables.append({"name" : name, "value" : value})    
    
    def get_register_job_definition_args(self, job_name, step):
        self.set_container_variables(step)
        job_def_args = {
            'jobDefinitionName': job_name,
            "type": "container",
            "containerProperties": {
                "image": self.scar_batch_io_image_id,
                "vcpus": 1,
                "memory": self.lambda_instance.memory,                       
                "command": ["scar-batch-io"],
                "volumes": [
                    {"host": {
                        "sourcePath": self.lambda_instance.input_folder},
                     "name": "SCAR_INPUT_DIR"},
                    {"host":{
                        "sourcePath": self.lambda_instance.output_folder},
                     "name": "SCAR_OUTPUT_DIR"},
                ],
                "environment" : self.container_environment_variables,                             
                'mountPoints': [
                    {"sourceVolume": "SCAR_INPUT_DIR",
                     "containerPath": self.lambda_instance.input_folder},
                    {"sourceVolume": "SCAR_OUTPUT_DIR",
                     "containerPath": self.lambda_instance.output_folder},
                ],
            },
        }
        if step == "MED":
            job_def_args["containerProperties"]["command"] = []
            job_def_args["containerProperties"]["image"] = utils.get_environment_variable("IMAGE_ID")
            if self.script != "":
                job_def_args["containerProperties"]["command"] = ["{0}/script.sh".format(self.lambda_instance.input_folder)]
        
        return job_def_args
    
    def register_job_definition(self, job_name, step):
        register_job_args = self.get_register_job_definition_args(job_name, step)
        self.client.register_job_definition(**register_job_args)
        
    def invoke_batch_function(self):
        # Register lambda Job
        self.register_job_definition(self.lambda_instance.function_name, "MED")
        # Submit download input Job
        job_id = None
        if self.lambda_instance.has_input_bucket() or self.script != "":
            self.register_job_definition(self.io_job_name, "INIT")
            job_id = self.submit_init_job()
        # Submit lambda Job
        lambda_job_id = self.submit_lambda_job(job_id)
        # Submit store output Job
        if self.lambda_instance.has_input_bucket() or self.lambda_instance.has_output_bucket():
            self.register_job_definition(self.io_job_name, "END")
            self.submit_end_job(lambda_job_id)
        return lambda_job_id
    
    def get_user_script(self):
        script = ""
        if utils.is_variable_in_environment('INIT_SCRIPT_PATH'):
            file_content = utils.read_file(utils.get_environment_variable('INIT_SCRIPT_PATH'), file_mode='rb')
            script = utils.utf8_to_base64_string(file_content)        
        if utils.is_value_in_dict(self.lambda_instance.event, 'script'):
            script = self.lambda_instance.event['script']
        return script
    
    def get_job_env_vars(self, step):
        variables= []
        scar_input_file = "" if not self.scar_input_file else self.scar_input_file
        self.add_environment_variable("STEP", step)
        self.add_environment_variable("SCRIPT", self.get_user_script())
        self.add_environment_variable("FUNCTION_NAME", self.lambda_instance.function_name)
        self.add_environment_variable("SCAR_INPUT_FILE", scar_input_file)
        self.add_environment_variable("SCAR_INPUT_DIR", self.lambda_instance.input_folder)
        self.add_environment_variable("SCAR_OUTPUT_DIR", self.lambda_instance.output_folder)
        self.add_environment_variable("REQUEST_ID", self.lambda_instance.request_id)

        if self.lambda_instance.has_input_bucket():
            self.add_environment_variable("INPUT_BUCKET", self.lambda_instance.input_bucket)
        if self.lambda_instance.has_output_bucket():
            self.add_environment_variable("OUTPUT_BUCKET", self.lambda_instance.output_bucket)
        
        for user_var, value in utils.get_user_defined_variables().items():
            variables.append({"name" : user_var, "value" : value})
        return variables
    
    def get_job_args(self, step, job_id=None):
        job_name =  self.lambda_instance.function_name if step == 'MED' else self.io_job_name
        job_def = {"jobDefinition" : job_name,
                   "jobName" : job_name,
                   "jobQueue" : self.lambda_instance.function_name,
                   "containerOverrides" : {
                       "environment" : self.get_job_env_vars(step)
                    }
                  }
        if job_id:
            job_def['dependsOn'] = [{'jobId' : job_id, 'type' : 'SEQUENTIAL'}]
        return job_def    
    
    def submit_batch_job(self, job_args):
        return self.client.submit_job(**job_args)["jobId"]
    
    def submit_init_job(self):
        return self.submit_batch_job(self.get_job_args('INIT'))
    
    def submit_lambda_job(self, job_id):
        return self.submit_batch_job(self.get_job_args('MED', job_id))
    
    def submit_end_job(self, job_id):
        return self.submit_batch_job(self.get_job_args('END', job_id))



# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import boto3
import os
from urllib.parse import unquote_plus
import faassupervisor.utils as utils
from faassupervisor.interfaces.dataprovider import DataProviderInterface

logger = utils.get_logger()

class Minio(DataProviderInterface):

    @utils.lazy_property
    def client(self):
        client = boto3.client('s3', endpoint_url='http://minio-service.minio:9000',
                              aws_access_key_id=os.environ['AWS_ACCESS_KEY_ID'],
                              aws_secret_access_key=os.environ['AWS_SECRET_ACCESS_KEY'])
        return client

    def __init__(self, event, output_folder):
        self.event = event
        self.os_tmp_folder = utils.get_tmp_dir()
        self.output_folder = output_folder
    
    @classmethod    
    def is_minio_event(cls, event):
        if utils.is_key_and_value_in_dictionary('data', event) \
        and utils.is_key_and_value_in_dictionary('body', event['data']) \
        and utils.is_key_and_value_in_dictionary('Records', event['data']['body']):
            record = event['data']['body']['Records'][0]
            return record['eventSource'] == 'minio:s3'
        return False
    
    def get_record(self):
        return self.event['data']['body']['Records'][0]['s3']    
    
    def download_input(self):
        '''Downloads the file from the minio bucket and returns the path were the download is placed'''
        record_info = self.get_record()
        bucket = record_info['bucket']['name']
        key = unquote_plus(record_info['object']['key'])
        file_name = os.path.splitext(key)[0]
        file_download_path = "{0}/{1}".format(self.os_tmp_folder, file_name) 
        print("Downloading item from bucket '{0}' with key '{1}'".format(bucket, key))
        utils.create_folder(self.os_tmp_folder)
        with open(file_download_path, 'wb') as data:
            self.client.download_fileobj(bucket, key, data)
        print("Successful download of file '{0}' from bucket '{1}' in path '{2}'".format(key, bucket, file_download_path))
        return file_download_path
    
    def upload_output(self):
        output_files_path = utils.get_all_files_in_directory(self.output_folder)
        output_bucket = os.environ['OUTPUT_BUCKET']
        print("UPLOADING FILES {0}".format(output_files_path))
        for file_path in output_files_path:
            file_name = file_path.replace("{0}/".format(self.output_folder), "")
            output_file_name = "{0}-out{1}".format(os.path.splitext(file_name)[0],''.join(os.path.splitext(file_name)[1:]))
            self.upload_file(output_bucket, file_path, output_file_name)
    
    def upload_file(self, bucket_name, file_path, file_key):
        print("Uploading file  '{0}' to bucket '{1}'".format(file_key, bucket_name))
        with open(file_path, 'rb') as data:
            self.client.upload_fileobj(data, bucket_name, file_key)



# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import boto3
import os
from urllib.parse import unquote_plus
import faassupervisor.utils as utils
from faassupervisor.interfaces.dataprovider import DataProviderInterface

logger = utils.get_logger()

class S3(DataProviderInterface):
    
    @utils.lazy_property
    def client(self):
        client = boto3.client('s3')
        return client
    
    def __init__(self, lambda_instance):
        self.lambda_instance = lambda_instance
        if utils.is_value_in_dict(self.lambda_instance.event, 'Records'):
            self.record = self.get_s3_record()
            self.input_bucket = self.record['bucket']['name']
            self.file_key = unquote_plus(self.record['object']['key'])
            self.file_name = os.path.basename(self.file_key).replace(' ', '')
            self.file_download_path = '{0}/{1}'.format(self.lambda_instance.input_folder, self.file_name)

    @classmethod
    def is_s3_event(cls, event):
        if utils.is_value_in_dict(event, 'Records'):
            return event['Records'][0]['eventSource'] == "aws:s3"
        return False

    def get_s3_record(self):
        if len(self.lambda_instance.event['Records']) > 1:
            logger.warning("Multiple records detected. Only processing the first one.")
            
        record = self.lambda_instance.event['Records'][0]
        if utils.is_value_in_dict(record, 's3'):
            return record['s3']

    def download_input(self):
        '''Downloads the file from the S3 bucket and returns the path were the download is placed'''
        logger.info("Downloading item from bucket '{0}' with key '{1}'".format(self.input_bucket, self.file_key))
        if not os.path.isdir(self.file_download_path):
            os.makedirs(os.path.dirname(self.file_download_path), exist_ok=True)
        with open(self.file_download_path, 'wb') as data:
            self.client.download_fileobj(self.input_bucket, self.file_key, data)
        logger.info("Successful download of file '{0}' from bucket '{1}' in path '{2}'".format(self.file_key, 
                                                                                               self.input_bucket,
                                                                                               self.file_download_path))
        return self.file_download_path
  
    def get_file_key(self, function_name=None, folder=None, file_name=None):
        if function_name:
            return "{0}/{1}/{2}/{3}".format(function_name, folder, self.lambda_instance.request_id, file_name)
        else:
            return "{0}/{1}/{2}".format(folder, self.lambda_instance.request_id, file_name)

    def upload_output(self, bucket_name, bucket_folder=None):
        output_files_path = utils.get_all_files_in_directory(self.lambda_instance.output_folder)
        logger.debug("UPLOADING FILES {0}".format(output_files_path))
        for file_path in output_files_path:
            file_name = file_path.replace("{0}/".format(self.lambda_instance.output_folder), "")
            if bucket_folder:
                file_key = self.get_file_key(folder=bucket_folder, file_name=file_name)
            else:
                file_key = self.get_file_key(function_name=self.lambda_instance.function_name, folder='output', file_name=file_name)
            self.upload_file(bucket_name, file_path, file_key)
            
    def upload_file(self, bucket_name, file_path, file_key):
        logger.info("Uploading file  '{0}' to bucket '{1}'".format(file_key, bucket_name))
        with open(file_path, 'rb') as data:
            self.client.upload_fileobj(data, bucket_name, file_key)
        logger.info("Changing ACLs for public-read for object in bucket {0} with key {1}".format(bucket_name, file_key))
        obj = boto3.resource('s3').Object(bucket_name, file_key)
        obj.Acl().put(ACL='public-read')
    
    def download_file_to_memory(self, bucket_name, file_key):
        obj = boto3.resource('s3').Object(bucket_name=bucket_name, key=file_key)
        print ("Reading item from bucket {0} with key {1}".format(bucket_name, file_key))
        return obj.get()["Body"].read()
    
    def delete_file(self):
        self.client.delete_object(Bucket=self.input_bucket, Key=self.file_key)
        


# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import faassupervisor.utils as utils
logger = utils.get_logger()

class Lambda():

    def __init__(self, event, context):
        self.event = event
        self.context = context
        self.request_id = context.aws_request_id
        self.memory = int(context.memory_limit_in_mb)
        self.arn = context.invoked_function_arn
        self.function_name = context.function_name
        self.log_group_name = self.context.log_group_name
        self.log_stream_name = self.context.log_stream_name  
        self.permanent_folder = "/var/task"
        self._set_tmp_folders()         

    def _set_tmp_folders(self):
        self.temporal_folder = utils.create_tmp_dir()
        self.temporal_folder_path = self.temporal_folder.name
        self.input_folder = "{}/input".format(self.temporal_folder_path)
        self.output_folder = "{}/output".format(self.temporal_folder_path)            

    @utils.lazy_property
    def output_bucket(self):
        output_bucket = utils.get_environment_variable('OUTPUT_BUCKET')
        return output_bucket
    
    @utils.lazy_property
    def output_bucket_folder(self):
        output_bucket_folder = utils.get_environment_variable('OUTPUT_FOLDER')
        return output_bucket_folder
    
    @utils.lazy_property
    def input_bucket(self):
        input_bucket = utils.get_environment_variable('INPUT_BUCKET')
        return input_bucket
    
    def has_output_bucket(self):
        return utils.is_variable_in_environment('OUTPUT_BUCKET')

    def has_output_bucket_folder(self):
        return utils.is_variable_in_environment('OUTPUT_FOLDER')
    
    def has_input_bucket(self):
        return utils.is_variable_in_environment('INPUT_BUCKET')

    def get_invocation_remaining_seconds(self):
        return int(self.context.get_remaining_time_in_millis() / 1000) - int(utils.get_environment_variable('TIMEOUT_THRESHOLD'))
    



# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import uuid
import base64
import faassupervisor.utils as utils
logger = utils.get_logger()

class ApiGateway():

    def __init__(self, lambda_instance):
        self.lambda_instance = lambda_instance

    def is_post_request_with_body(self):
        return self.lambda_instance.event['httpMethod'] == 'POST'

    def is_post_request_with_body_json(self):
        return self.lambda_instance.event['httpMethod'] == 'POST' and self.lambda_instance.event['headers']['Content-Type'].strip() == 'application/json'

    def is_request_with_parameters(self):
        return"queryStringParameters" in self.lambda_instance.event and self.lambda_instance.event["queryStringParameters"]

    def save_post_body_json(self):
        body = self.lambda_instance.event['body']
        file_path = "/tmp/{0}/api_event.json".format(self.lambda_instance.request_id)
        logger.info("Received JSON from POST request and saved it in path '{0}'".format(file_path))
        self.save_file(file_path, 'w', body)            
        return file_path

    def save_post_body_file(self):
        body = base64.b64decode(self.lambda_instance.event['body'])
        body_file_name = uuid.uuid4().hex
        file_path = "/tmp/{0}/{1}".format(self.lambda_instance.request_id, body_file_name)
        logger.info("Received file from POST request and saved it in path '{0}'".format(file_path))
        self.save_file(file_path, 'wb', body)
        return file_path
    
    def save_file(self, file_path, write_mode, content):
        os.makedirs(os.path.dirname(file_path), exist_ok=True)  
        with open(file_path, write_mode) as data:
            data.write(content)

    def save_post_body(self):
        if 'body' in self.lambda_instance.event and self.lambda_instance.event['body']:
            if self.is_post_request_with_body_json():
                file_path = self.save_post_body_json()
            elif self.is_post_request_with_body:
                file_path = self.save_post_body_file()
            return file_path
        
    def save_request_parameters(self):
        if self.is_request_with_parameters():
            self.lambda_instance.http_params = {}
            for key, value in self.lambda_instance.event["queryStringParameters"].items():
                self.lambda_instance.http_params[format(key)] = value




# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import subprocess
import traceback
import faassupervisor.utils as utils
from faassupervisor.exceptions import NoStorageProviderDefinedWarning
from faassupervisor.interfaces.supervisor import SupervisorInterface
from faassupervisor.providers.aws.lambda_.function import Lambda
from faassupervisor.providers.aws.lambda_.udocker import Udocker
from faassupervisor.providers.aws.batch.batch import Batch
from faassupervisor.providers.aws.apigateway.apigateway import ApiGateway
from faassupervisor.providers.aws.storage.s3 import S3

logger = utils.get_logger()

class LambdaSupervisor(SupervisorInterface):
    
    @utils.lazy_property
    def batch(self):
        batch = Batch(self.lambda_instance, self.scar_input_file)
        return batch
    
    @utils.lazy_property
    def apigateway(self):
        apigateway = ApiGateway(self.lambda_instance)
        return apigateway
    
    @utils.lazy_property
    def udocker(self):
        udocker = Udocker(self.lambda_instance, self.scar_input_file)
        return udocker
    
    @utils.lazy_property
    def storage_client(self):
        if S3.is_s3_event(self.lambda_instance.event):
            storage_client = S3(self.lambda_instance)
        else:
            raise NoStorageProviderDefinedWarning()
        return storage_client 
    
    def __init__(self, **kwargs):
        logger.info('SUPERVISOR: Initializing AWS Lambda supervisor')
        logger.debug("EVENT: {}".format(kwargs['event']))
        logger.debug("CONTEXT: {}".format(kwargs['context']))
        self.lambda_instance = Lambda(kwargs['event'], kwargs['context'])
        self.create_temporal_folders()
        self.create_event_file()
        self.status_code = 200
        self.body = {}
        self.scar_input_file = None

    def is_apigateway_event(self):
        return 'httpMethod' in self.lambda_instance.event           
           
    def prepare_udocker(self):
        self.udocker.create_image()
        self.udocker.create_container()
        self.udocker.create_command()
        
    def execute_udocker(self):
        try:
            udocker_output = self.udocker.launch_udocker_container()
            logger.info("CONTAINER OUTPUT:\n " + udocker_output)
            self.body["udocker_output"] = udocker_output            
        except subprocess.TimeoutExpired:
            logger.warning("Container execution timed out")
            if(utils.get_environment_variable("EXECUTION_MODE") == "lambda-batch"):
                self.execute_batch()
                
    def has_input_bucket(self):
        return hasattr(self, "input_bucket") and self.input_bucket and self.input_bucket != "" 

    def upload_to_bucket(self):
        bucket_name = None
        bucket_folder = None
        
        if self.lambda_instance.has_output_bucket():
            bucket_name = self.lambda_instance.output_bucket
            logger.debug("OUTPUT BUCKET SET TO {0}".format(bucket_name))
            if self.lambda_instance.has_output_bucket_folder():
                bucket_folder = self.lambda_instance.output_bucket_folder
                logger.debug("OUTPUT FOLDER SET TO {0}".format(bucket_folder))
                
        elif self.lambda_instance.has_input_bucket():
            bucket_name = self.lambda_instance.input_bucket
            logger.debug("OUTPUT BUCKET SET TO {0}".format(bucket_name))
            
        if bucket_name:
            self.storage_client.upload_output(bucket_name, bucket_folder)

    def create_temporal_folders(self):
        utils.create_folder(self.lambda_instance.input_folder)
        utils.create_folder(self.lambda_instance.output_folder)
        utils.create_folder(utils.get_environment_variable("UDOCKER_DIR"))
    
    def create_event_file(self):
        utils.create_file_with_content("{0}/event.json".format(self.lambda_instance.temporal_folder_path), json.dumps(self.lambda_instance.event))        

    def execute_batch(self):
        batch_ri = self.batch.invoke_batch_function()
        batch_logs = "Check batch logs with: \n  scar log -n {0} -ri {1}".format(self.lambda_instance.function_name, batch_ri)
        self.body["udocker_output"] = "Job delegated to batch.\n{0}".format(batch_logs)         

    def is_batch_execution(self):
        return utils.get_environment_variable("EXECUTION_MODE") == "batch"
    
    ##################################################################
    ## The methods below must be defined for the supervisor to work ##
    ##################################################################
    
    def parse_input(self):
        if  S3.is_s3_event(self.lambda_instance.event):
            self.input_bucket = self.storage_client.input_bucket
            logger.debug("INPUT BUCKET SET TO {0}".format(self.input_bucket))
            if self.is_batch_execution():
                self.scar_input_file = self.storage_client.file_download_path
            else:
                self.scar_input_file = self.storage_client.download_input()
            logger.debug("INPUT FILE SET TO {0}".format(self.scar_input_file))
        elif self.is_apigateway_event():
            self.apigateway.save_request_parameters()
            self.scar_input_file = self.apigateway.save_post_body()
            logger.debug("INPUT FILE SET TO {0}".format(self.scar_input_file)) 
    
    def parse_output(self):
        self.upload_to_bucket()
    
    def execute_function(self):
        if self.is_batch_execution():
            self.execute_batch()
        else:
            self.prepare_udocker()
            self.execute_udocker()
            
    def create_error_response(self):
        exception_msg = traceback.format_exc()
        logger.error("Exception launched:\n {0}".format(exception_msg))
        return {"statusCode" : 500, 
                "headers" : { 
                    "amz-lambda-request-id": self.lambda_instance.request_id, 
                    "amz-log-group-name": self.lambda_instance.log_group_name, 
                    "amz-log-stream-name": self.lambda_instance.log_stream_name },
                "body" : json.dumps({"exception" : exception_msg}),
                "isBase64Encoded" : False                
                }

    def create_response(self):
        return {"statusCode" : self.status_code, 
                "headers" : { 
                    "amz-lambda-request-id": self.lambda_instance.request_id, 
                    "amz-log-group-name": self.lambda_instance.log_group_name, 
                    "amz-log-stream-name": self.lambda_instance.log_stream_name },
                "body" : json.dumps(self.body),
                "isBase64Encoded" : False }             



# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import base64
import json
import os
import re
import subprocess
import tarfile
import tempfile
import uuid
import sys
import platform
import shutil
from faassupervisor.exceptions import InvalidPlatformError
import logging

def resource_path(relative_path, bin_path=None):
    """ Get absolute path to resource, works for dev and for PyInstaller """
    try:
        # PyInstaller creates a temp folder and stores path in _MEIPASS
        base_path = sys._MEIPASS
    except Exception:
        if bin_path:
            return bin_path
        else:
            base_path = os.path.abspath(".")
    return os.path.join(base_path, relative_path)

def is_binary_execution():
    try:
        binary_env = sys._MEIPASS
        if platform.system().lower() != 'linux':
            raise InvalidPlatformError()
        return True
    except Exception:
        return False

def get_logger():
    logger = logging.getLogger()
    if is_variable_in_environment('LOG_LEVEL'):
        logger.setLevel(get_environment_variable('LOG_LEVEL'))
    else:
        logger.setLevel('INFO')
    return logger

def join_paths(*paths):
    return os.path.join(*paths)

def get_tmp_dir():
    return tempfile.gettempdir()

def create_tmp_dir():
    return tempfile.TemporaryDirectory()

def get_random_tmp_folder():
    return join_paths(tempfile.gettempdir(), str(uuid.uuid4().hex))    

def lazy_property(func):
    ''' A decorator that makes a property lazy-evaluated.'''
    attr_name = '_lazy_' + func.__name__

    @property
    def _lazy_property(self):
        if not hasattr(self, attr_name):
            setattr(self, attr_name, func(self))
        return getattr(self, attr_name)
    return _lazy_property

def find_expression(string_to_search, rgx_pattern):
    '''Returns the first group that matches the rgx_pattern in the string_to_search'''
    if string_to_search:    
        pattern = re.compile(rgx_pattern)
        match = pattern.search(string_to_search)
        if match :
            return match.group()

def base64_to_utf8_string(value):
    return base64.b64decode(value).decode('utf-8')

def utf8_to_base64_string(value):
    return base64.b64encode(value).decode('utf-8')

def dict_to_base64_string(value):
    return base64.b64encode(json.dumps(value)).decode("utf-8")

def divide_list_in_chunks(elements, chunk_size):
    """Yield successive n-sized chunks from th elements list."""
    if len(elements) == 0:
        yield []
    for i in range(0, len(elements), chunk_size):
        yield elements[i:i + chunk_size]
        
def get_random_uuid4_str():
    return str(uuid.uuid4())

def merge_dicts(d1, d2):
    '''
    Merge 'd1' and 'd2' dicts into 'd1'.
    'd2' has precedence over 'd1'
    '''
    for k,v in d2.items():
        if v:
            if k not in d1:
                d1[k] = v
            elif type(v) is dict:
                d1[k] = merge_dicts(d1[k], v)
            elif type(v) is list:
                d1[k] += v
    return d1

def is_value_in_dict(dictionary, value):
    return value in dictionary and dictionary[value]

def get_tree_size(path):
    """Return total size of files in given path and subdirs."""
    total = 0
    for entry in os.scandir(path):
        if entry.is_dir(follow_symlinks=False):
            total += get_tree_size(entry.path)
        else:
            total += entry.stat(follow_symlinks=False).st_size
    return total

def get_all_files_in_directory(dir_path):
    files = []
    for dirname, _, filenames in os.walk(dir_path):
        for filename in filenames:
            files.append(os.path.join(dirname, filename))
    return files

def get_file_size(file_path):
    '''Return file size in bytes'''
    return os.stat(file_path).st_size

def create_folder(folder_name):
    if not os.path.isdir(folder_name):
        os.makedirs(folder_name, exist_ok=True)
        
def create_file_with_content(path, content):
    with open(path, "w") as f:
        f.write(content)

def read_file(file_path, file_mode="r", file_encoding="utf-8"):
    if file_mode == 'rb':
        file_encoding = None
    with open(file_path, mode=file_mode, encoding=file_encoding) as content_file:
        return content_file.read()
    
def delete_file(path):
    if os.path.isfile(path):
        os.remove(path)

def delete_folder(path):
    shutil.rmtree(path)
    
def create_tar_gz(files_to_archive, destination_tar_path):
    with tarfile.open(destination_tar_path, "w:gz") as tar:
        for file_path in files_to_archive:
            tar.add(file_path, arcname=os.path.basename(file_path))
    return destination_tar_path
      
def extract_tar_gz(tar_path, destination_path):
    with tarfile.open(tar_path, "r:gz") as tar:
        tar.extractall(path=destination_path)

def kill_process(self, process):
    # Using SIGKILL instead of SIGTERM to ensure the process finalization 
    os.killpg(os.getpgid(process.pid), subprocess.signal.SIGKILL)

def execute_command(command):
    subprocess.call(command)
    
def execute_command_and_return_output(command):
    return subprocess.check_output(command).decode("utf-8")

def is_variable_in_environment(variable):
    return is_value_in_dict(os.environ, variable)

def is_key_and_value_in_dictionary(key, dictionary):
    return (key in dictionary) and dictionary[key] and dictionary[key] != ""

def set_environment_variable(key, variable):
    if key and variable:
        os.environ[key] = variable

def get_environment_variable(variable):
    if is_variable_in_environment(variable):
        return os.environ[variable]

def parse_arg_list(arg_keys, cmd_args):
    result = {}
    for key in arg_keys:
        if type(key) is tuple:
            if key[0] in cmd_args and cmd_args[key[0]]:
                result[key[1]] = cmd_args[key[0]]
        else:
            if key in cmd_args and cmd_args[key]:
                result[key] = cmd_args[key]
    return result

def get_user_defined_variables():
    user_vars = {}
    for key in os.environ.keys():
        # Find global variables with the specified prefix
        if re.match("CONT_VAR_.*", key):
            user_vars[key.replace("CONT_VAR_", "")] = get_environment_variable(key)
    return user_vars

def get_stdin():
    buf = ""
    for line in sys.stdin:
        buf = buf + line
    return buf

def set_file_execution_rights(file_path):
    execute_command(["chmod", "+x", file_path])



# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import faassupervisor.utils as utils
import faassupervisor.exceptions as excp
from faassupervisor.interfaces.supervisor import SupervisorInterface
from faassupervisor.providers.aws.batch.job import BatchJob
from faassupervisor.providers.aws.storage.s3 import S3

logger = utils.get_logger()

class BatchSupervisor(SupervisorInterface):
    
    @utils.lazy_property
    def storage_client(self):
        if S3.is_s3_event(self.batch_job.event):
            storage_client = S3(self.batch_job)
        else:
            raise excp.NoStorageProviderDefinedWarning()
        return storage_client     
    
    def __init__(self, **kwargs):
        logger.info('SUPERVISOR: Initializing AWS Batch supervisor')
        logger.debug("EVENT: {}".format(kwargs['event']))
        logger.debug("CONTEXT: {}".format(kwargs['context']))
        self.batch_job = BatchJob(kwargs['event'], kwargs['context'])

    def create_user_script(self):
        if utils.is_variable_in_environment('SCRIPT'):
            script_path = utils.join_paths(self.batch_job.input_folder, 'script.sh')
            script_content = utils.base64_to_utf8_string(utils.get_environment_variable('SCRIPT'))
            utils.create_file_with_content(script_path, script_content)
            logger.info("Script file created in '{0}'".format(script_path))
            utils.set_file_execution_rights(script_path)
     
    def upload_to_bucket(self):
        bucket_name = None
        bucket_folder = None
    
        if self.batch_job.has_output_bucket():
            bucket_name = self.batch_job.output_bucket
            logger.info("OUTPUT BUCKET SET TO {0}".format(bucket_name))
    
            if self.batch_job.has_output_bucket_folder():
                bucket_folder = self.batch_job.output_bucket_folder
                logger.info("OUTPUT FOLDER SET TO {0}".format(bucket_folder))
    
        elif self.batch_job.has_input_bucket():
            bucket_name = self.batch_job.input_bucket
            logger.info("OUTPUT BUCKET SET TO {0}".format(bucket_name))
    
        if bucket_name:
            self.storage_client.upload_output(bucket_name, bucket_folder)

    ##################################################################
    ## The methods below must be defined for the supervisor to work ##
    ##################################################################

    # Not needed in BATCH execution
    def execute_function(self):
        pass    
    
    @excp.exception(logger)    
    def parse_input(self):
        step = utils.get_environment_variable("STEP")
        if step == "INIT":
            logger.info("INIT STEP")
            self.create_user_script()
            if utils.is_variable_in_environment('INPUT_BUCKET'):
                self.storage_client.download_input()  
    
    @excp.exception(logger)
    def parse_output(self):
        step = utils.get_environment_variable("STEP")
        if step == "END":
            logger.info("END STEP")
            self.upload_to_bucket()
        
    def create_response(self):
        pass
    
    def create_error_response(self):
        pass



# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import faassupervisor.utils as utils
logger = utils.get_logger()

# Executed inside the containers launched in each batch job
class BatchJob():
    
    def __init__(self, event, context):
        self.event = event
        self._set_context_info(context)
        self._set_tmp_folders()         
        
    def _set_context_info(self, context):
        self.request_id = context['aws_request_id']
        self.memory = int(context['memory_limit_in_mb'])
        self.function_name = context['function_name']
        self.log_group_name = context['log_group_name']
        self.log_stream_name = context['log_stream_name']        

    def _set_tmp_folders(self):
        self.input_folder = utils.get_environment_variable('SCAR_INPUT_DIR')
        self.output_folder = utils.get_environment_variable('SCAR_OUTPUT_DIR')        

    @utils.lazy_property
    def output_bucket(self):
        output_bucket = utils.get_environment_variable('OUTPUT_BUCKET')
        return output_bucket
    
    @utils.lazy_property
    def output_bucket_folder(self):
        output_bucket_folder = utils.get_environment_variable('OUTPUT_FOLDER')
        return output_bucket_folder
    
    @utils.lazy_property
    def input_bucket(self):
        input_bucket = utils.get_environment_variable('INPUT_BUCKET')
        return input_bucket
    
    def has_output_bucket(self):
        return utils.is_variable_in_environment('OUTPUT_BUCKET')

    def has_output_bucket_folder(self):
        return utils.is_variable_in_environment('OUTPUT_FOLDER')
    
    def has_input_bucket(self):
        return utils.is_variable_in_environment('INPUT_BUCKET')
    